/* Copyright (c) 2009 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase Media Stream for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  Created by Barry Leslie on 3/20/09.
 *
 */
 
#include "CSConfig.h"

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#ifdef SUPPORT_CLOUDS
#include "libs3.h"
#endif

#include "CSGlobal.h"
#include "CSThread.h"
#include "CSLog.h"
#include "CSPath.h"
#include "CSFile.h"
#include "CSString.h"
#include "CSStrUtil.h"
#include "CSStorage.h"
#include "CSEncode.h"

#include "cloud.h"

#ifdef SUPPORT_CLOUDS
static S3UriStyle s3_UriStyle = S3UriStyleVirtualHost;

class KeyItem : public CSRefObject {
private:
	/* Make this object linkable: */
	CSObject		*iNextLink;
	CSObject		*iPrevLink;
	void setNextLink(CSObject *link) { iNextLink = link; }
	void setPrevLink(CSObject *link) { iPrevLink = link; }
	
	char *key_buffer;

public:
	CSObject *getNextLink() { return iNextLink; }
	CSObject *getPrevLink() { return iPrevLink; }

	KeyItem(const char *str)
	{
		iNextLink = iPrevLink = NULL;
		key_buffer = (char *) cs_malloc(strlen(str) + 1);
		strcpy(key_buffer, str);
	}
	
	~KeyItem()
	{
		if (key_buffer)
			cs_free(key_buffer);
	}
	
	const char *getCString() { return key_buffer;}
	
};


class s3CallBackData {
	public:
	s3CallBackData(CloudDB *data): cloud(data){}
	CloudDB *cloud;
};

class s3PutCallBackData : public s3CallBackData {
	public:
	CSFile *inFile;
	off_t pos;
	off_t len;
	s3PutCallBackData(CloudDB *cloud, CSFile *input, off_t offset, off_t size): s3CallBackData(cloud),inFile(input), pos(offset), len(size) {}
};

class s3GetCallBackData : public s3CallBackData {
	public:
	char *buffer;
	off_t size;
	bool overflow;
	s3GetCallBackData(CloudDB *cloud, char *buff, off_t bsize): s3CallBackData(cloud),buffer(buff), size(bsize), overflow(false) {}
};

class s3ListCallBackData : public s3CallBackData {
	public:
	s3ListCallBackData(CloudDB *cloud, CSLinkedList *list): s3CallBackData(cloud), key_list(list), isTruncated(0), prefix(cloud->clObjectKey->getCString()), prefix_length(cloud->clObjectKeyBaseLength){}
	CSLinkedList *key_list;

	char *prefix;
	int prefix_length;
    int isTruncated;
    char nextMarker[1024];
};


//////////////////////////////
// Cload info file functions:
//////////////////////////////
#ifndef PBMS_CLOUD_ENCRYPT_KEY
#define PBMS_CLOUD_ENCRYPT_KEY "PBMS_CLOUD"
#endif

#define MAX_INFO_LINE (MAX_NODE_SIZE + MAX_PUBLIC_KEY + MAX_PRIVATE_KEY + 100) // leave lots of room for whitespace.
#define MIN_INFO_LINE 5

/*-----------------------------------------------------------*/
static csWord4 hash(const char *str)
{
	csWord4 v = 0;
	while (*str) {
		v += (v +1) * *str;
		str++;
	}
	return v;
}

/*-----------------------------------------------------------*/
/*
 * This is just to prevent someone from being able to findout what the key
 * is just by reading the file.
 * The real security is up to the user to make sure the key file is secure.
 */
static void encryptCloudInfo(CSFile *info_file, CloudKey *key)
{
	csWord4 i;
	csWord4 *ptr, anum, len=0;
	CloudKey target = *key;
	enter_();
	push_(info_file);
	

	ptr = (csWord4 *)&target;
	len = sizeof(CloudKey)/sizeof(*ptr);
	
	anum = hash(PBMS_CLOUD_ENCRYPT_KEY);
	
	/* Fill the target with rubish */
	for (i = 0; i < len; i++, ptr++) *ptr = rand();

	strcpy(target.node, key->node);
	strcpy(target.public_key, key->public_key);
	strcpy(target.private_key, key->private_key);
	ptr = (csWord4 *)&target;

	for (i = 0; i < len; i++, ptr++) *ptr = (*ptr + anum) ^ anum;

	info_file->setEOF(0);
	info_file->setEOF(0);
	
	info_file->write("#", 0, 1);
	info_file->write(&target, 1, sizeof(CloudKey));
	
	release_(info_file);
	exit_();
}

/*-----------------------------------------------------------*/
static void unencryptCloudInfo(char *encrypt_key, CloudKey *info)
{
	csWord4 i;
	csWord4 *ptr, anum, len=0;
	
	ptr = (csWord4 *)info;
	len = sizeof(CloudKey)/sizeof(*ptr);
	memcpy(ptr, encrypt_key+1, sizeof(CloudKey));
	anum = hash(PBMS_CLOUD_ENCRYPT_KEY);
	
	for (i = 0; i < len; i++, ptr++) *ptr = (*ptr ^ anum) - anum;
	
}

/*-----------------------------------------------------------*/
static void loadCloudInfo(CSFile *info_file, CloudKey *info)
{
	int cnt;
	char line[MAX_INFO_LINE], *ptr, *privkey, *pubkey, *node;
	const char  *err = NULL;
	enter_();
	push_(info_file);
	
	cnt = info_file->read(line, 0, MAX_INFO_LINE, MIN_INFO_LINE);
	line[cnt] = 0;
	
	if (*line == '#') {
		unencryptCloudInfo(line, info);
		goto done;
	}
	
	err = 	"End of line found before Cloud node name";
	
	ptr = line;
	while (isspace(*ptr)) ptr++;	
	if (!*ptr) {
		goto error;
	}
	
	node = ptr;
	while (*ptr && !isspace(*ptr)) ptr++;
	if (!*ptr) 
		goto error;
	*ptr = 0;
	ptr++;
	
	//-------------
	err = 	"End of line found before Public Key";
	while (isspace(*ptr)) ptr++;	
	if (!*ptr) {
		goto error;
	}
	
	pubkey = ptr;
	while (*ptr && !isspace(*ptr)) ptr++;
	if (!*ptr) 
		goto error;
	*ptr = 0;
	ptr++;
	
	//-------------
	err = 	"End of line found before Private Key";
	while (isspace(*ptr)) ptr++;	
	if (!*ptr) 
		goto error;
	
	privkey = ptr;
	while (*ptr && !isspace(*ptr)) ptr++;
	if (*ptr) {
		*ptr = 0;
		ptr++;
		while (isspace(*ptr)) ptr++;
		if (*ptr && (ptr - line) < cnt) {	
			err = 	"Unexpected chatrater(s) after private key";
			goto error;
		}
	}
	
	if (strlen(pubkey) >= MAX_PUBLIC_KEY || strlen(privkey) >= MAX_PRIVATE_KEY || strlen(node) >= MAX_NODE_SIZE ) {
		err = "Maximum key/node length exceeded.";
		goto error;
	}
	
	strcpy(info->node, node);
	strcpy(info->public_key, pubkey);
	strcpy(info->private_key, privkey);
	
	info_file->retain();
	encryptCloudInfo(info_file, info);
	goto done;

error:
	{
		char msg[1024];
		cs_strcpy(1024, msg, info_file->getPathString());
		cs_strcat(1024, msg, ": ERROR:  ");
		cs_strcat(1024, msg, err);
		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, msg);
	}
done:	
	release_(info_file);
	exit_();
	
}

////////////////////////////
/////////////////////////////


static void handle_S3_status(S3Status s3_status)
{
	enter_();
	switch (s3_status) {
		case S3StatusOK:
			break;
		
		case S3StatusOutOfMemory:
			CSException::throwOSError(CS_CONTEXT, ENOMEM);
		break;
			
		default:
			CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, S3_get_status_name(s3_status));
	}
	
	exit_();
}

// This callback does the same thing for every request type: prints out the
// properties if the user has requested them to be so
#define print_nonnull(p, name, field)		\
    do {									\
        if (properties-> field) {			\
			p->append(name);				\
			p->append(": ");				\
			p->append(properties->field);	\
			p->append("\n");				\
        }									\
    } while (0)
  
#define SKIP_DUMP
static S3Status responsePropertiesCallback(const S3ResponseProperties *properties, void *callbackData)
{
	CloudDB *data = ((s3CallBackData*) callbackData)->cloud;
	
    if (!data->tracing) 
        return S3StatusOK;

#ifdef SKIP_DUMP
	return S3StatusOK;
#else
	enter_();
	CSStringBuffer *prop_info;
	new_(prop_info, CSStringBuffer());	
			
	push_(prop_info);

    print_nonnull(prop_info, "Content-Type", contentType);
    print_nonnull(prop_info, "Request-Id", requestId);
    print_nonnull(prop_info, "Request-Id-2", requestId2);
    if (properties->contentLength > 0) {
		prop_info->append("Content-Length: ");
		prop_info->append((u_llong) properties->contentLength);
		prop_info->append("\n");
    }
    print_nonnull(prop_info, "Server", server);
    print_nonnull(prop_info, "ETag", eTag);
	
    if (properties->lastModified > 0) {
        char timebuf[256];
        time_t t = (time_t) properties->lastModified;
		struct tm	ltime;
 		localtime_r(&t, &ltime);

        strftime(timebuf, sizeof(timebuf), "%Y-%m-%dT%H:%M:%SZ", &ltime);
 		prop_info->append("Last-Modified: ");
		prop_info->append(timebuf);
		prop_info->append("\n");
   }
    for (int i = 0; i < properties->metaDataCount; i++) {
		prop_info->append("x-amz-meta-");
		prop_info->append(properties->metaData[i].name);
		prop_info->append(": ");
		prop_info->append(properties->metaData[i].value);
		prop_info->append("\n");
    }

	release_(prop_info);
    return_(S3StatusOK);
#endif
}


//-----------------------------------------------------------
static void responseCompleteCallback(S3Status status, const S3ErrorDetails *error, void *callbackData)
{
	CloudDB *data = ((s3CallBackData*) callbackData)->cloud;

	data->status = status;
	if (status == S3StatusOK)
		return;
		
	data->can_retry = S3_status_is_retryable(status);
	data->file_not_found = (status == S3StatusErrorNoSuchKey) || (status == S3StatusHttpErrorNotFound) ;
	
	if (error && !data->throw_error) {
		enter_();
		try_(a) {
			CSStringBuffer *err;
			new_(err, CSStringBuffer());			
			push_(err);
			
			err->append("AmazonS3:");
			if (error->message) {
				err->append("Message: ");
				err->append(error->message);
			} else {
				err->append("S3 Status: ");
				err->append(S3_get_status_name(status));
			}
			
			if (error->resource) {
				err->append(" Resource: ");
				err->append(error->resource);
			}
			
			if (error->furtherDetails) {
				err->append(" Further Details: ");
				err->append(error->furtherDetails);
			}
			
			if (error->extraDetailsCount) {
				err->append(" Extra Details: ");
				for (int i = 0; i < error->extraDetailsCount; i++) {
					if (i)
						err->append(", ");
					err->append(error->extraDetails[i].name);
					err->append(": ");
					err->append(error->extraDetails[i].value);					
				}
			}
			
			CSException::RecordException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, err->getCString());
			release_(err);
		}
		
		catch_(a);
		cont_(a);
		data->throw_error = true;
		exit_();
	}
	
}

//==============================
CloudDB::CloudDB(CloudDB *cloud)
{
	enter_();
	push_(cloud);
	
	keep_alive = cloud->keep_alive;
	memcpy(&keys, &(cloud->keys), sizeof(CloudKey));
	
	new_(clObjectKey, CSStringBuffer());
	
	cloud->cl_clearObjectKey();
	clObjectKey->append(cloud->clObjectKey->getCString());
	clObjectKeyBaseLength = clObjectKey->length();


	release_(cloud);
	exit_();
}

//==============================
CloudDB::CloudDB(CSFile *cloud_info_file, csWord4 db_id, time_t url_keep_alive)
{
	S3Status s3_status;
	enter_();

	keep_alive = url_keep_alive;
	new_(clObjectKey, CSStringBuffer());

	// All keys are prefixed with "<db_id>/"
	clObjectKey->append((int)db_id);
	clObjectKey->append("/");
	
	clObjectKeyBaseLength = clObjectKey->length();
	
	loadCloudInfo(cloud_info_file, &keys);
		
	s3_status = S3_initialize(NULL, S3_INIT_ALL);
	handle_S3_status(s3_status);
	exit_();
}

//==============================
CloudDB::~CloudDB()
{
	
	if (clObjectKey)
		clObjectKey->release();
	S3_deinitialize();
}

//==============================
CloudDB *CloudDB::cl_clone()
{
	CloudDB *cdb;	
	enter_();
	retain();
	new_(cdb, CloudDB(this));
	return_(cdb);
}

//==============================
void CloudDB::cl_create_db()
{
// This is a noop. 
}

//-----------------------------------------------------------
static S3Status listBucketCallback(int isTruncated, const char *nextMarker,
                                   int contentsCount, 
                                   const S3ListBucketContent *contents,
                                   int commonPrefixesCount,
                                   const char **commonPrefixes,
                                   void *callbackData)
{
	s3ListCallBackData *data = (s3ListCallBackData *) callbackData;    
	S3Status status = S3StatusOK;
	
	enter_();

    data->isTruncated = isTruncated;
    // This is tricky.  S3 doesn't return the NextMarker if there is no
    // delimiter.  Why, I don't know, since it's still useful for paging
    // through results.  We want NextMarker to be the last content in the
    // list, so set it to that if necessary.
    if ((!nextMarker || !nextMarker[0]) && contentsCount) {
        nextMarker = contents[contentsCount - 1].key;
    }
    
     if (nextMarker) {
        snprintf(data->nextMarker, sizeof(data->nextMarker), "%s", nextMarker);
    }
    else {
        data->nextMarker[0] = 0;
    }
    
	try_(a) {
		for (int i = 0; i < contentsCount; i++) {
			const char *key = contents[i].key;
			
			if (! memcmp(key, data->prefix, data->prefix_length)) {
				KeyItem *item;
				new_(item, KeyItem(key + data->prefix_length));
				data->key_list->addFront(item);
			}
		}
	}
		
	catch_(a) {
		data->cloud->throw_error = true;
		status = S3StatusInternalError;
	}
	
	cont_(a);

     return_(status);
}

//==============================
// Drop database deletes all objects with the database key prefix
void CloudDB::cl_drop_db()
{
	CSLinkedList *list;
	KeyItem *key;
	cl_clearObjectKey();
	s3ListCallBackData callback_data(this, list); // Must call cl_clearObjectKey() before setting this.
    S3BucketContext bucketContext = { keys.node, S3ProtocolHTTPS, s3_UriStyle, keys.public_key, keys.private_key };

	S3ListBucketHandler listBucketHandler = { { &responsePropertiesCallback, &responseCompleteCallback }, &listBucketCallback };
	callback_data.nextMarker[0] = 0;
	enter_();
	
	new_(list, CSLinkedList());
	push_(list);
	
	// Get a list of the keys in te database.
	unsigned long  retries;
	do {
        callback_data.isTruncated = 0;
		retries = max_retries;
        do {
			if (retries != max_retries)
				self->sleep(sleep_time);
				
			clearState();				
            S3_list_bucket(&bucketContext, clObjectKey->getCString(), (callback_data.nextMarker[0])?callback_data.nextMarker: NULL, NULL, 0, 0, &listBucketHandler, &callback_data);
			retries--;
        } while (can_retry && retries);
		
		if (throw_error) 
			throw_();
			
    } while (callback_data.isTruncated);


	// Go through the list deleting the keys.
	while ((key = (KeyItem*)(list->removeBack())) ) {
		push_(key);
		cl_deleteData(key->getCString());
		release_(key);
	}
	
	release_(list);
	exit_();
}

//==============================
bool CloudDB::cl_db_exists()
{
	bool found = false;
 	s3CallBackData callbackData(this);
	
	enter_();
    S3ResponseHandler responseHandler = { &responsePropertiesCallback, &responseCompleteCallback };
	csWord4  retries = max_retries;


    do {
		if (retries != max_retries)
			self->sleep(sleep_time);
		clearState();
        S3_test_bucket(S3ProtocolHTTPS, s3_UriStyle, keys.public_key, keys.private_key,
                         keys.node, 0, NULL, NULL,
                         &responseHandler, &callbackData);
 		retries--;
    } while (can_retry && retries);

	if (status == S3StatusOK)
		found = true;
		
	if (throw_error) 
		throw_();

	return_(found);
}

//==============================
static int putObjectDataCallback(int bufferSize, char *buffer, void *callbackData)
{
	s3PutCallBackData *data = (s3PutCallBackData *) callbackData;    
	int toRead = ((data->len > bufferSize) ? bufferSize : data->len);
	enter_();

	if (toRead) {
		try_(a) {
			data->inFile->read(buffer, data->pos, toRead, toRead);
			data->pos += toRead;
			data->len -= toRead;
		}
		catch_(a);
		toRead = 0;
		data->cloud->throw_error = true;
		cont_(a);		
	}

    return_(toRead);
}

//-----------------------------------------------------------
void CloudDB::cl_putData(const char * key, CSFile *inFile, off_t offset, off_t size, Md5Digest *md5)
{
	s3PutCallBackData callback_data(this, inFile, offset, size);
	char *md5Encoded = NULL;


	enter_();

	push_(inFile);
	
	cl_setObjectKey(key);
	md5Encoded = base64Encode(md5->val, 16); // Do not forget to free this!

    S3BucketContext bucketContext = { keys.node, S3ProtocolHTTPS, s3_UriStyle, keys.public_key, keys.private_key };
    S3PutProperties putProperties = {
        NULL,				// contentType
        md5Encoded,
        NULL,				// cacheControl
        NULL,				// contentDispositionFilename
        NULL,				// contentEncoding
        0,					// expires
        S3CannedAclPrivate,	//cannedAcl
        0,					//metaPropertiesCount
        NULL				// metaProperties
    };

    S3PutObjectHandler putObjectHandler = { { &responsePropertiesCallback, &responseCompleteCallback }, &putObjectDataCallback };

	
	unsigned long  retries = max_retries;
    do {
		callback_data.pos = offset;
		callback_data.len = size;
		if (retries != max_retries) 
			self->sleep(sleep_time);
		
		clearState();				
        S3_put_object(&bucketContext, clObjectKey->getCString(), size, &putProperties, NULL, &putObjectHandler, &callback_data);
					  
		retries--;
    } while (can_retry && retries);
	
	
	cs_free(md5Encoded);
	if (throw_error)
		throw_();
	release_(inFile);	
	exit_();
}

//==============================
static S3Status getObjectDataCallback(int bufferSize, const char *buffer, void *callbackData)
{
	s3GetCallBackData *data = (s3GetCallBackData *) callbackData;    

	// In an overflow condition we just continue 
	// inorder to find out how much data there actually was.
	if (data->overflow)
		data->size += bufferSize;
	else if (data->size < bufferSize) { 
		data->overflow = true;
		bufferSize -= data->size;
		data->size = bufferSize;
	} else {
		memcpy(data->buffer, buffer, bufferSize);
		data->buffer += bufferSize;
		data->size -= bufferSize;
	}
    return(S3StatusOK);
}

//-----------------------------------------------------------
off_t CloudDB::cl_getData(const char * key, char *buffer, off_t size)
{
	s3GetCallBackData callback_data(this, buffer, size);
	enter_();

    S3BucketContext bucketContext = { keys.node, S3ProtocolHTTPS, s3_UriStyle, keys.public_key, keys.private_key };
    S3GetConditions getConditions =
    {
        -1,		// ifModifiedSince,
        -1,		// ifNotModifiedSince,
        NULL,	// ifMatch,		
        NULL	// ifNotMatch
    };
   S3GetObjectHandler getObjectHandler = { { &responsePropertiesCallback, &responseCompleteCallback }, &getObjectDataCallback };

	
	cl_setObjectKey(key);
	unsigned long  retries = max_retries;
    do {
		callback_data.buffer = buffer;
		callback_data.size = size;
		if (retries != max_retries) 
			self->sleep(sleep_time);
		
		clearState();				
        S3_get_object(&bucketContext, clObjectKey->getCString(), &getConditions, 0, 0, NULL, &getObjectHandler, &callback_data);
					  
		retries--;
    } while (can_retry && retries);
	
	
	if (throw_error)
		throw_();
		
	if (callback_data.overflow) {
		CSStringBuffer *err;
		new_(err, CSStringBuffer());			
		push_(err);
		err->append("AmazonS3: Buffer overflow, database: ");
		err->append(keys.node);
		err->append(" key: ");
		err->append(key);
		err->append(" expected size: ");
		err->append((int)size);
		err->append(" data recieved: ");
		err->append((int)(size + callback_data.size));
		
		CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, err->getCString());
	}
	
	return_(size - callback_data.size);
}

//==============================
void CloudDB::cl_deleteData(const char *key)
{
	s3CallBackData callback_data(this);
	enter_();
	
    S3BucketContext bucketContext = { keys.node, S3ProtocolHTTPS, s3_UriStyle, keys.public_key, keys.private_key };
 
    S3ResponseHandler responseHandler = { &responsePropertiesCallback, &responseCompleteCallback };

	cl_setObjectKey(key);
	unsigned long  retries = max_retries;
    do {
		if (retries != max_retries)
			self->sleep(sleep_time);
		clearState();				
        S3_delete_object(&bucketContext, clObjectKey->getCString(),  NULL, &responseHandler, &callback_data);

		retries--;
    } while (can_retry && retries);

			
	if (throw_error) 
		throw_();
		
	exit_();
}

//==============================
CSString *CloudDB::cl_getDataURL(const char *key)
{
	S3Status status;
	char query[S3_MAX_AUTHENTICATED_QUERY_STRING_SIZE];
	enter_();
	
	clearState();
	cl_setObjectKey(key);
		
	memset(query, 0, S3_MAX_AUTHENTICATED_QUERY_STRING_SIZE); // There is a bug in S3_generate_authenticated_query_string() that it doesn't null terminate the string.
    S3BucketContext bucketContext = { keys.node, S3ProtocolHTTP, s3_UriStyle, keys.public_key, keys.private_key };
	 
	if ((status = S3_generate_authenticated_query_string(query, &bucketContext, clObjectKey->getCString(), keep_alive + time(NULL), NULL)) != S3StatusOK) {
		handle_S3_status(status);
	}
	
printf("\n===========================\n");
printf("-->%s<--", query);
printf("\n===========================\n");
	CSString *url;
	
	url = CSString::newString(query);
	return_(url);
}
#else // SUPPORT_CLOUDS
//==============================
CloudDB *CloudDB::cl_clone() { return NULL;}
void CloudDB::cl_create_db(){}
void CloudDB::cl_drop_db(){}
bool CloudDB::cl_db_exists(){return false;}

void CloudDB::cl_putData(const char *key, CSFile *inFile, off_t f_offset, off_t size, Md5Digest *md5){}
off_t CloudDB::cl_getData(const char *key, char *data, off_t size){return 0;}
CSString *CloudDB::cl_getDataURL(const char *key){return NULL;}
void CloudDB::cl_deleteData(const char *key){}
#endif // SUPPORT_CLOUDS

